<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Admin\Models;

use Carbon\Carbon;
use Illuminate\Database\Eloquent\Model;
use MicroEngine\Admin\Scopes\UserScope;
use MicroEngine\Core\Pagination\PaginationTrait;

class User extends Model
{
    use PaginationTrait;

    /**
     * @var array
     */
    protected $fillable = ['login_id', 'password', 'name', 'email', 'portlets', 'bookmarks', 'note'];
    /**
     * 配列に含めない属性
     * @var array
     */
    protected $hidden = ['password'];
    /**
     * @var array
     */
    protected $casts = [
        'portlets' => 'array',
        'bookmarks' => 'array',
    ];

    /**
     * モデルの「初期起動」メソッド
     *
     * @return void
     */
    protected static function boot()
    {
        parent::boot();

        static::addGlobalScope(new UserScope());
    }

    /**
     * ユーザーのブックマークを取得
     *
     * @return array
     */
    public function getBookmarks(): array
    {
        if ($this->bookmarks === null || !is_array($this->bookmarks)) {
            return [];
        }
        return $this->bookmarks;
    }

    /**
     * ロック状態確認
     *
     * @return bool
     */
    public function isLocked(): bool
    {
        // 設定取得
        $lockoutLimit = config('admin.lockout_limit', 10);
        $lockoutDuration = config('admin.lockout_duration', 30);

        // 失敗回数が制限値に達している場合
        if ($this->failure_count >= $lockoutLimit && $this->failed_at !== null) {
            // 最後の失敗日時からロック期間分未来の日時をロック解除日時とする
            $unlockedAt = Carbon::parse($this->failed_at)->addMinutes($lockoutDuration);
            if (Carbon::now()->lt($unlockedAt)) {
                // 現在日時より、ロック解除日時が未来の場合はロック状態と判断する
                return true;
            }
        }
        return false;
    }

    /**
     * ロック期間が経過したかどうかを確認
     *
     * @return bool
     */
    public function lockPeriodElapsed(): bool
    {
        // 設定取得
        $lockoutLimit = config('admin.lockout_limit', 10);
        $lockoutDuration = config('admin.lockout_duration', 30);

        // 失敗が制限値に到達している
        if ($this->failure_count >= $lockoutLimit && $this->failed_at !== null) {
            // 最後の失敗日時からロック期間分未来の日時をロック解除日時とする
            $unlockedAt = Carbon::parse($this->failed_at)->addMinutes($lockoutDuration);
            if (Carbon::now()->gte($unlockedAt)) {
                // ロック解除日時が過去の場合はロック期間が過ぎたと判定する
                return true;
            }
        }
        return false;
    }
}
