<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Calendar;

use Carbon\Carbon;
use MicroEngine\Cms\Models\Calendar;

class MonthlyCalendar
{
    /**
     * @var string
     */
    protected $date;
    /**
     * @var array
     */
    protected $config;
    /**
     * @var array
     */
    protected $days;
    /**
     * @var array
     */
    protected $daysOfWeek;

    /**
     * DBからデータを取り出してプロパティに格納する
     *
     * @param string $type
     * @param string|null $date
     * @param int $amount
     * @param bool $published
     * @throws \Exception
     */
    public function pull(string $type, string $date = null, int $amount = 1, bool $published = true)
    {
        $this->config = content_config($type, 'calendar');
        $this->config += [
            'first_day_of_week' => 0,
        ];
        $firstDayOfWeek = $this->config['first_day_of_week'];

        $this->date = $this->monthlyTargetDate($date);
        $this->days = $this->range($type, $this->date, $amount, 0, $published);
        $this->daysOfWeek = $this->daysOfWeek($firstDayOfWeek);
    }

    /**
     * 月別データを返す
     *
     * @param bool $sevenDays
     * @return array
     */
    public function monthly(bool $sevenDays = true): array
    {
        $months = [];

        foreach ($this->days as $date => $day) {
            $firstDate = substr($date, 0, 7) . '-01';
            if (!isset($months[$firstDate])) {
                $months[$firstDate] = [];
            }
            $months[$firstDate][] = $day;
        }

        if ($sevenDays) {
            // 7日（１週間）ずつのデータに別ける
            foreach ($months as $key => $days) {
                $months[$key] = $this->convertWeekly($days);
            }
        }
        return $months;
    }

    /**
     * 日別データを返す
     *
     * @return array
     */
    public function daily(): array
    {
        return $this->days;
    }

    /**
     * @return array
     */
    public function getConfig(): array
    {
        return $this->config;
    }

    /**
     * @return array
     */
    public function getDaysOfWeek(): array
    {
        return $this->daysOfWeek;
    }

    /**
     * 範囲指定でデータを取得
     *
     * @param string $type
     * @param string $date
     * @param int $monthPeriod
     * @param int $dayPeriod
     * @param bool $published
     * @return array
     */
    protected function range(string $type, string $date, int $monthPeriod = 1, int $dayPeriod = 0, bool $published = true): array
    {
        $days = [];

        // クラス名用の曜日
        $daysOfWeek = ['sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat'];
        $today = date('Y-m-d');

        // 開始日・終了日を定義
        $startDt = Carbon::parse($date);
        $endDt = $startDt->copy()->addMonths($monthPeriod)->addDays($dayPeriod);

        // DBから該当範囲のレコードを取得
        $query = Calendar::where('type', $type);
        if (!$published) {
            // 管理画面の場合は論理削除以外を取得
            $query->notDeleted();
        }
        $models = $query
            ->where('date', '>=', $startDt->format('Y-m-d'))
            ->where('date', '<', $endDt->format('Y-m-d'))
            ->orderBy('date')
            ->get();

        // 範囲の日付データを作成
        while ($startDt->lt($endDt)) {
            $year = $startDt->format('Y');
            $month = $startDt->format('m');
            $day = $startDt->format('d');
            $dateString = $year . '-' . $month . '-' . $day;
            $days[$dateString] = [
                'date' => $dateString,
                'year' => $year,
                'month' => $month,
                'day' => $day,
                'day_of_week' => $daysOfWeek[$startDt->dayOfWeek],
                'today' => ($today === $dateString),
                'content' => null,
            ];
            // 1日進める
            $startDt->addDay();
        }

        // DBから取得したレコードを日付データにセットする。
        foreach ($models as $model) {
            $days[$model->date]['content'] = $model;
        }
        return $days;
    }

    /**
     * 月間カレンダーの対象日付を返す
     *
     * @param string|null $date
     * @return string
     */
    protected function monthlyTargetDate(string $date = null): string
    {
        if ($date === null) {
            // null の場合は今月１日をセットする
            $date = date('Y-m-01');
        } elseif (is_numeric($date)) {
            // 数値の場合は今月１日を基準に指定月数を前後させる
            $date = Carbon::today()->addMonths($date)->format('Y-m-01');
        } else {
            // 1日をセットする
            $date = Carbon::parse($date)->format('Y-m-01');
        }
        return $date;
    }

    /**
     * 7日ごとのデータに変換する
     *
     * @param array $days
     * @return array
     */
    protected function convertWeekly(array $days): array
    {
        $weeks = [];
        $week = [];
        $firstDayOfWeek = $this->config['first_day_of_week'];
        $i = 0;
        foreach ($days as $day) {
            // 1日の場合
            if ($day['day'] === '01') {
                $dt = Carbon::parse($day['date']);
                // 曜日の開始日から１日のオフセットを確認
                $offset = $dt->dayOfWeek - $firstDayOfWeek;
                // オフセットがマイナスの場合は１つ手前の行に戻っているので、7 を足す。
                if ($offset < 0) {
                    $offset += 7;
                }
                // オフセット分空配列を追加する
                $week = array_pad($week, $offset, []);
                // オフセット分カウンターを進める
                $i += $offset;
            }
            $week[] = $day;
            $i++;

            // ７列目まで到達した場合
            if ($i === 7) {
                $i = 0;
                $weeks[] = $week;
                $week = [];
            }
        }
        // 最終週の処理
        if ($i) {
            $week = array_pad($week, 7, []);
            $weeks[] = $week;
        }
        return $weeks;
    }

    /**
     * 曜日情報を返す
     *
     * @param int $firstDayOfWeek
     * @return array
     */
    protected function daysOfWeek(int $firstDayOfWeek = 0): array
    {
        $daysOfWeek = [];

        // 曜日の開始日を調整する
        if (!is_numeric($firstDayOfWeek) || $firstDayOfWeek < 0 || $firstDayOfWeek > 7) {
            $firstDayOfWeek = 0;
        }
        $keyList = ['sun', 'mon', 'tue', 'wed', 'thu', 'fri', 'sat'];

        // 曜日の表示文字列を取得する
        $valueList = config('cms.days_of_week', ['日', '月', '火', '水', '木', '金', '土']);

        foreach (range($firstDayOfWeek, 6) as $key) {
            $daysOfWeek[$keyList[$key]] = $valueList[$key];
        }
        if ($firstDayOfWeek) {
            foreach (range(0, ($firstDayOfWeek - 1)) as $key) {
                $daysOfWeek[$keyList[$key]] = $valueList[$key];
            }
        }
        return $daysOfWeek;
    }
}
