<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Controllers;

use MicroEngine\Cms\Models\Page;
use MicroEngine\Cms\Forms\PageForm;
use Psr\Http\Message\ResponseInterface;
use Interop\Container\ContainerInterface;
use MicroEngine\Cms\Services\PageManager;
use Psr\Http\Message\ServerRequestInterface;
use MicroEngine\Core\Controllers\Controller;
use MicroEngine\Cms\Services\ContentService;
use MicroEngine\Cms\Services\CategoryManager;

class PageController extends Controller
{
    /**
     * @var PageManager
     */
    protected $pageManager;
    /**
     * @var CategoryManager
     */
    protected $categoryManager;

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        parent::__construct($c);

        $this->pageManager = $c->pageManager;
        $this->categoryManager = $c->categoryManager;
        // ライセンスチェック
        $this->c->license->hasLicense();
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function index(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $pagesConfig = config('pages');
        history_unset('page_info');
        history_unset('page_edit');

        return $this->view->render(
            $response,
            'cms/views/page/index.twig',
            compact('pagesConfig')
        );
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function listing(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        history_set('page_info', $request);
        history_set('page_edit', $request);
        $type = $args['type'];
        $pageNo = $args['page_no'] ?? 1;
        $sortArg = $args['sort'] ?? 'updated_at';
        $sort = valid_parameter($sortArg, ['updated_at', 'name', 'path']);
        $orderArg = $args['order'] ?? 'desc';
        $order = valid_parameter($orderArg, ['desc', 'asc']);
        $config = content_config($type, 'page');

        $pagination = Page::notDeleted()->where('type', $type)->orderBy($sort, $order)->pagination($pageNo);
        $pagination->setRouteInfo($this->c, 'admin.page.listing', ['type' => $type, 'sort' => $sort, 'order' => $order]);
        $sortOrder = $this->c->util->sortOrderClass($sort, $order, ['name', 'path', 'updated_at']);

        return $this->view->render(
            $response,
            'cms/views/page/listing.twig',
            compact('type', 'config', 'pagination', 'sortOrder')
        );
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function create(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $type = $args['type'];
        $config = content_config($type, 'page');
        $categories = $this->categoryManager->categoryOptions($config);
        $form = new PageForm($this->c);
        $form->setConfig($config);
        $form->setInputValue([
            'category_id' => $request->getQueryParam('category'),
            'filename' => $this->pageManager->initialFilename($config),
        ]);
        $form->setInitialValues();
        $form->set('edit_mode', 'create');

        return $this->view->render(
            $response,
            'cms/views/page/create.twig',
            compact('categories', 'type', 'config', 'form')
        );
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function copy(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $page = Page::withoutGlobalScopes()->findOrFail($args['id']);
        $type = $page->type;
        $config = content_config($type, 'page');
        $categories = $this->categoryManager->categoryOptions($config);
        unset($page->id);
        if (isset($config['inputs']['filename']['value'])) {
            $page->filename = $this->pageManager->initialFilename($config);
        }
        $form = new PageForm($this->c);
        $form->setConfig($config);
        $form->setInputValue($page);
        $form->set('edit_mode', 'copy');

        return $this->view->render(
            $response,
            'cms/views/page/create.twig',
            compact('categories', 'type', 'config', 'form')
        );
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function store(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $type = $args['type'];
        $config = content_config($type, 'page');
        $form = new PageForm($this->c);
        $form->setConfig($config);
        $form->setInputValue($request->getParsedBody());

        // バリデーション
        if ($form->validate()) {
            $values = $form->getInputValues();
            $page = $this->pageManager->create($values);
            if ($form->get('revision_type') === 'draft') {
                $this->flash->addMessage('success', "ページを下書きとして保存しました。名前：{$page->name} ( {$page->path} )");
            } else {
                $this->flash->addMessage('success', "ページを作成しました。名前：{$page->name} ( {$page->path} )");
            }

            if ($form->get('edit_mode') === 'copy') {
                // 複製した場合はhistoryに関係なく新ページの情報画面に遷移させる。
                return $response->withRedirect($this->c->router->pathFor('admin.page.info', ['id' => $page->id]));
            }
            if (history_has('page_edit')) {
                return $response->withRedirect(history('page_edit'));
            }
            return $response->withRedirect($this->c->router->pathFor('admin.page.listing', ['type' => $page->type]));
        } else {
            $categories = $this->categoryManager->categoryOptions($config);

            return $this->view->render(
                $response,
                'cms/views/page/create.twig',
                compact('categories', 'type', 'config', 'form')
            );
        }
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function preview(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $type = $args['type'];
        $config = content_config($type, 'page');
        $form = new PageForm($this->c);
        $form->setConfig($config);
        $form->setInputValue($request->getParsedBody());
        $site = $this->c->site;
        $site->setPreview(true);

        // バリデーション
        if ($form->validate()) {
            $values = $form->getInputValues(false);
            $page = new Page($values);
            $site->pushContent($page, true);

            return $this->view->render(
                $response,
                $page->getTemplatePath(),
                compact('page')
            );
        } else {
            return $this->view->render(
                $response,
                'cms/views/base/previewError.twig',
                compact('form')
            );
        }
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function edit(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $id = $args['id'] ?? null;
        $editMode = $args['edit_mode'] ?? 'edit';
        $revisionId = $args['revision_id'] ?? null;
        $open = $request->getQueryParam('open') ?? 'form';

        if ($open === 'preview') {
            // 公開側から遷移してきた場合はヒストリーをリセットする
            history_unset('page_info');
            history_unset('page_edit');
        }

        if ($revisionId) {
            $page = $this->pageManager->revision($id, $revisionId);
        } else {
            $page = Page::withoutGlobalScopes()->findOrFail($id);
        }
        // 公開設定が-5の場合は、ごみ箱に入ったコンテンツなので、編集モードを trash にする。
        if ($page->visibility == ContentService::VISIBILITY_TRASH) {
            $editMode = 'trash';
            // ごみ箱内のコンテンツは公開をセットする
            $page->visibility = ContentService::VISIBILITY_PUBLIC;
        }
        $type = $page->type;
        $config = content_config($type, 'page');
        $categories = $this->categoryManager->categoryOptions($config);
        $form = new PageForm($this->c);
        $form->setId($args['id']);
        $form->setConfig($config);
        $form->setInputValue($page);
        $form->set('edit_mode', $editMode);
        $form->set('revision_id', $revisionId);

        return $this->view->render(
            $response,
            'cms/views/page/edit.twig',
            compact('categories', 'type', 'config', 'form', 'open')
        );
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function update(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $page = Page::withoutGlobalScopes()->findOrFail($args['id']);
        $type = $page->type;
        $config = content_config($type, 'page');
        $form = new PageForm($this->c);
        $form->setId($args['id']);
        $form->setConfig($config);
        $form->setInputValue($request->getParsedBody());

        // バリデーション
        if ($form->validate()) {
            $values = $form->getInputValues();
            $page = $this->pageManager->update($args['id'], $values);
            if ($form->get('revision_type') === 'draft') {
                $this->flash->addMessage('success', "ページを下書きとして保存しました。名前：{$page->name} ( {$page->path} )");
            } else {
                $this->flash->addMessage('success', "ページを更新しました。名前：{$page->name} ( {$page->path} )");
            }

            if (history_has('page_edit')) {
                return $response->withRedirect(history('page_edit'));
            }
            return $response->withRedirect($this->c->router->pathFor('admin.page.listing', ['type' => $page->type]));
        } else {
            $categories = $this->categoryManager->categoryOptions($config);

            return $this->view->render(
                $response,
                'cms/views/page/edit.twig',
                compact('categories', 'type', 'config', 'form')
            );
        }
    }

    /**
     * @param ServerRequestInterface $request
     * @param ResponseInterface $response
     * @param array $args
     * @return ResponseInterface
     * @throws \Exception
     */
    public function info(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        history_set('page_edit', $request);

        $page = Page::withoutGlobalScopes()->findOrFail($args['id']);
        $type = $page->type;
        $config = content_config($type, 'page');
        $revisions = $this->pageManager->revisionList($page);

        return $this->view->render(
            $response,
            'cms/views/page/info.twig',
            compact('page', 'type', 'config', 'revisions')
        );
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function view(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $id = $args['id'] ?? null;
        $revisionId = $args['revision_id'] ?? null;
        if ($revisionId) {
            $page = $this->pageManager->revision($id, $revisionId);
        } else {
            $page = Page::withoutGlobalScopes()->findOrFail($id);
        }

        $site = $this->c->site;
        $site->pushContent($page, true);

        return $this->view->render(
            $response,
            $page->getTemplatePath(),
            compact('page')
        );
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function destroy(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $id = $args['id'];
        $editMode = $args['edit_mode'] ?? null;
        $revisionId = $args['revision_id'] ?? null;

        $page = $this->pageManager->destroy($id, $revisionId);
        if ($editMode === 'revision') {
            $message = '履歴データを削除しました。';
        } elseif ($editMode === 'draft') {
            $message = '下書きを削除しました。';
        } else {
            $message = 'ページをごみ箱に移動しました。';
        }
        $this->flash->addMessage('success', $message . "名前：{$page->name} ( {$page->path} )");

        if ($editMode !== null && history_has('page_edit')) {
            return $response->withRedirect(history('page_edit'));
        }
        return $response->withRedirect($this->c->router->pathFor('admin.page.listing', ['type' => $page->type]));
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function trash(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        // ごみ箱内のコンテンツを取得
        $pages = Page::trashed()->orderBy('path')->get();

        return $this->view->render(
            $response,
            'cms/views/page/trash.twig',
            compact('pages')
        );
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function forceDestroy(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        $id = $args['id'];

        $page = $this->pageManager->forceDestroy($id);
        $this->flash->addMessage('success', "ページをごみ箱から削除しました。名前：{$page->name} ( {$page->path} )");

        return $response->withRedirect($this->c->router->pathFor('admin.page.trash'));
    }

    /**
     * @param  ServerRequestInterface $request
     * @param  ResponseInterface $response
     * @param  array $args
     * @return ResponseInterface
     */
    public function emptyTrash(ServerRequestInterface $request, ResponseInterface $response, array $args): ResponseInterface
    {
        // ごみ箱を空にして、削除したコンテンツ数を取得
        $count = $this->pageManager->emptyTrash();
        $this->flash->addMessage('success', 'ごみ箱を空にしました。削除ページ数：' . $count);

        return $response->withRedirect($this->c->router->pathFor('admin.page.trash'));
    }
}
