<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Managers;

use MicroEngine\Cms\Models\Collection;
use Interop\Container\ContainerInterface;
use MicroEngine\Cms\Services\ContentService;
use MicroEngine\Cms\Models\CollectionRevision;

class CollectionManager
{
    use RevisionTrait;

    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var string モデルクラス名
     */
    protected $modelClass = 'MicroEngine\Cms\Models\Collection';
    /**
     * @var string IDカラム名
     */
    protected $modelId = 'collection_id';
    /**
     * @var string リビジョンクラス名
     */
    protected $revisionClass = 'MicroEngine\Cms\Models\CollectionRevision';
    /**
     * @var string リビジョンテーブル名
     */
    protected $revisionTable = 'collection_revisions';

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
    }

    /**
     * コレクション保存
     *
     * @param array $data
     * @return Collection
     */
    public function create(array $data): Collection
    {
        // 作成者・更新者にログインユーザーのIDをセットする
        $data['create_user_id'] = $data['update_user_id'] = $this->c->auth->getUserId();

        $collection = Collection::create($data);
        $contentService = $this->c->contentService;
        $contentService->saveMeta($collection);
        $contentService->saveSearch($collection);

        // リビジョン保存
        $revision = $this->saveRevision($collection->id, $data);

        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            // 下書きとして保存
            // リビジョンIDを下書きIDとしてセット
            $collection->draft_id = $revision->id;
            // 新規で下書き保存した場合には、本稿が存在しないので、下書きだけというステータスを入れる
            $collection->visibility = ContentService::VISIBILITY_DRAFT;
            $collection->save();
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('コレクションを作成しました。', [
            'type' => $collection->type,
            'name' => $collection->name,
        ]);
        return $collection;
    }

    /**
     * コレクション更新
     *
     * @param int $id
     * @param array $data
     * @return Collection
     */
    public function update(int $id, array $data): Collection
    {
        // 更新者にログインユーザーのIDをセットする
        $data['update_user_id'] = $this->c->auth->getUserId();

        // DBからコレクションを取得
        $collection = Collection::withoutGlobalScopes()->findOrFail($id);

        // リビジョン保存
        $revision = $this->saveRevision($id, $data);

        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            // 下書きとして保存
            // リビジョンIDを下書きIDとしてセット
            $collection->draft_id = $revision->id;
            $collection->save();
        } else {
            // 本稿として保存
            $collection->fill($data);
            // 下書き編集画面から本稿登録の場合は下書きを削除する
            if ($data['edit_mode'] === 'draft') {
                $collection->draft_id = null;
            }
            $collection->save();

            $contentService = $this->c->contentService;
            $contentService->saveMeta($collection);
            $contentService->saveSearch($collection);
        }

        // ログを書き出す
        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            $message = 'コレクションを下書きとして保存しました。';
        } else {
            $message = 'コレクションを更新しました。';
        }
        $this->c->logger->addInfo($message, [
            'type' => $collection->type,
            'name' => $collection->name,
        ]);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $collection;
    }

    /**
     * コレクション削除
     *
     * @param int $id
     * @param int|null $revisionId
     * @return Collection
     */
    public function destroy(int $id, int $revisionId = null): Collection
    {
        $collection = Collection::withoutGlobalScopes()->findOrFail($id);
        // 更新者にログインユーザーのIDをセットする
        $collection->update_user_id = $this->c->auth->getUserId();

        // 履歴から再編集している場合。ただし、新規下書き保存状態の場合をのぞく。
        if ($revisionId && $collection->visibility != ContentService::VISIBILITY_DRAFT) {
            // 履歴から再編集している場合は、履歴を削除する。
            $revision = CollectionRevision::findOrFail($revisionId);
            $revision->delete();
            // 削除対象が最新下書きの場合はnullをセットする。
            if ($collection->draft_id == $revisionId) {
                $collection->draft_id = null;
                $collection->save();
                $message = 'コレクションの下書きを削除しました。';
            } else {
                $message = 'コレクションの履歴データを削除しました。';
            }

            // ログを書き出す
            $this->c->logger->addInfo($message, [
                'type' => $collection->type,
                'name' => $collection->name,
                'revision_id' => $revisionId,
            ]);
        } else {
            // 履歴から編集していない場合は、本レコードを論理削除する。
            $collection->visibility = ContentService::VISIBILITY_TRASH;
            $collection->update();

            // ログを書き出す
            $this->c->logger->addInfo('コレクションをゴミ箱に移動しました。', [
                'type' => $collection->type,
                'name' => $collection->name,
                'id' => $collection->id,
            ]);
        }
        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $collection;
    }

    /**
     * コレクションをごみ箱から削除
     *
     * @param int $id
     * @return Collection
     */
    public function forceDestroy(int $id): Collection
    {
        $collection = Collection::withoutGlobalScopes()->findOrFail($id);

        // ゴミ箱から削除する場合は、物理削除を行う
        $collection->delete();

        // ログを書き出す
        $this->c->logger->addInfo('コレクションをごみ箱から削除しました。', [
            'type' => $collection->type,
            'name' => $collection->name,
        ]);
        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $collection;
    }

    /**
     * ごみ箱を空にする
     *
     * @return int
     */
    public function emptyTrash(): int
    {
        $count = 0;
        // ごみ箱内の全コンテンツに対して処理する
        foreach (Collection::trashed()->get() as $collection) {
            // コンテンツ一つずつをごみ箱から削除する
            $this->forceDestroy($collection->id);
            $count++;
        }
        $this->c->logger->addInfo('コレクションのごみ箱を空にしました。', [
            'count' => $count,
        ]);
        return $count;
    }

    /**
     * 一括作成
     *
     * @param array $data
     * @return int $created 作成数
     */
    public function bulkCreate(array $data): int
    {
        $created = 0;

        foreach ($data as $row) {
            if (isset($row['fields']) && !is_array($row['fields'])) {
                $row['fields'] = json_decode($row['fields'], true);
            }
            $this->create($row);
            $created++;
        }
        $this->c->logger->addInfo('コレクションを一括作成しました。', [
            'count' => $created,
        ]);
        return $created;
    }
}
