<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Managers;

use MicroEngine\Cms\Models\File;
use MicroEngine\Core\Support\FileSystem;
use Interop\Container\ContainerInterface;
use Illuminate\Database\Eloquent\Collection;
use Illuminate\Database\Capsule\Manager as DB;

class FileManager
{
    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var FileSystem
     */
    protected $fileSystem;

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
        $filesDir = config('cms.files_dir', 'files');
        $this->fileSystem = new FileSystem(document_root($filesDir));
    }

    /**
     * ファイル保存
     *
     * @param array $data
     * @param bool $touch
     * @return File
     */
    public function create(array $data, bool $touch = true): File
    {
        // 作成者・更新者にログインユーザーのIDをセットする
        $data['create_user_id'] = $data['update_user_id'] = $this->c->auth->getUserId();

        $file = File::create($data);

        if ($touch) {
            $this->touchDir($file->dir);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ファイルを作成しました。', [
            'path' => $file->path(),
        ]);
        return $file;
    }

    /**
     * ファイル更新
     *
     * @param int $id
     * @param array $data
     * @param bool $touch
     * @return File
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function update(int $id, array $data, bool $touch = true): File
    {
        // 更新者にログインユーザーのIDをセットする
        $data['update_user_id'] = $this->c->auth->getUserId();

        $file = File::findOrFail($id);

        // 旧リサイズファイルを削除
        $this->destroyOldResize($file);

        $file->fill($data);
        $file->save();

        if ($touch) {
            $this->touchDir($file->dir);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ファイルを更新しました。', [
            'path' => $file->path(),
        ]);
        return $file;
    }

    /**
     * ファイル削除
     * @param int $id
     * @param bool $touch
     * @return File
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function destroy(int $id, bool $touch = true): File
    {
        $file = File::findOrFail($id);
        $this->fileSystem->delete($file->path());
        // 旧リサイズファイルを削除
        $this->destroyOldResize($file);

        $file->delete();

        if ($touch) {
            $this->touchDir($file->dir);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ファイルを削除しました。', [
            'path' => $file->path(),
        ]);
        return $file;
    }

    /**
     * ディレクトリ内のファイルコンテンツ取得
     *
     * @param string $dir
     * @param string|null $mode
     * @param string $sort
     * @param string $order
     * @return Collection
     */
    public function dirs(string $dir, string $mode = null, string $sort = 'filename', string $order = 'desc'): Collection
    {
        $query = File::query();
        if ($mode === 'image') {
            $query->where('dir', $dir)->where(function ($query) {
                $extensions = config('image.extensions');
                $query->whereIn('extension', $extensions)->orWhere('is_dir', true);
            });
        } else {
            $query->where('dir', $dir);
        }
        return $query->orderBy('is_dir', 'desc')->orderBy($sort, $order)->get();
    }

    /**
     * ディレクトリ作成
     *
     * @param array $data
     * @param bool $touch
     * @return File
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function createDir(array $data, bool $touch = true): File
    {
        // 作成者・更新者にログインユーザーのIDをセットする
        $data['create_user_id'] = $data['update_user_id'] = $this->c->auth->getUserId();

        $data['path'] = end_slash($data['dir']) . end_slash($data['filename']);
        $data['is_dir'] = true;
        $file = File::create($data);
        $this->fileSystem->makeDir($file->path());

        if ($touch) {
            $this->touchDir($file->dir);
        }
        $this->c->logger->addInfo('ディレクトリを作成しました。', [
            'path' => $file->path(),
        ]);
        return $file;
    }

    /**
     * ディレクトリ更新
     * @param int $id
     * @param array $data
     * @param bool $touch
     * @return File
     */
    public function updateDir(int $id, array $data, bool $touch = true): File
    {
        // 更新者にログインユーザーのIDをセットする
        $data['update_user_id'] = $this->c->auth->getUserId();

        $dir = File::findOrFail($id);
        $dir->fill($data);
        $dir->save();

        if ($touch) {
            $this->touchDir($dir->dir);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ディレクトリを更新しました。', [
            'path' => $dir->path(),
            'name' => $dir->name,
        ]);
        return $dir;
    }

    /**
     * ディレクトリ削除
     *
     * @param int $id
     * @param bool $touch
     * @return File
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function destroyDir(int $id, bool $touch = true): File
    {
        $dir = File::findOrFail($id);

        // ファイルシステム上のディレクトリを再帰的に削除する
        $this->fileSystem->deleteDir($dir->path());

        // 指定ディレクトリ内の全てのディレクトリとファイルを削除する
        DB::table('files')->where('dir', 'like', $dir->path() . '%')->delete();

        // 指定ディレクトリを削除する
        $dir->delete();

        if ($touch) {
            $this->touchDir($dir->dir);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ディレクトリを削除しました。', [
            'path' => $dir->path(),
            'name' => $dir->name,
        ]);
        return $dir;
    }

    /**
     * 現在年月期間のディレクトリを返す
     *
     * @return string
     */
    public function currentMonthDir(): string
    {
        return '/term/' . date('Y/m/');
    }

    /**
     * 現在年月のディレクトリを作成
     *
     * @return File
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function createCurrentDir(): File
    {
        $year = date('Y');
        $month = date('m');
        $dir = '/term/' . $year . '/';
        $file = File::where('dir', $dir)->where('filename', $month)->first();

        if (!$file) {
            // 期間ディレクトリの有無を確認
            if (!File::where('dir', '/')->where('filename', 'term')->count()) {
                // 無ければ作成する
                $this->createDir([
                    'dir' => '/',
                    'filename' => 'term',
                    'name' => '期間別',
                ]);
            }

            // 年ディレクトリの有無を確認
            if (!File::where('dir', '/term/')->where('filename', $year)->count()) {
                // 無ければ作成する
                $this->createDir([
                    'dir' => '/term/',
                    'filename' => $year,
                    'name' => $year . '年',
                ]);
            }

            // 今月のディレクトリを作成する
            $file = $this->createDir([
                'dir' => '/term/' . $year . '/',
                'filename' => $month,
                'name' => $year . '年' . $month . '月',
            ]);
            $this->c->logger->addInfo('ディレクトリを作成しました。', [
                'path' => $file->path(),
                'name' => $file->name,
            ]);
        }
        return $file;
    }

    /**
     * ディレクトリパス（パンくずナビ）用の配列を返す
     *
     * @param string $dir
     * @return array $paths
     */
    public function dirPaths(string $dir): array
    {
        // パンくずナビ配列
        $paths = [];

        $pieces = explode('/', trim($dir, '/'));
        $path = '/';
        foreach ($pieces as $piece) {
            $path .= $piece . '/';
            $paths[$path] = $piece;
        }
        return $paths;
    }

    /**
     * アップロードしたファイルを保存する
     *
     * @param array $uploadedFiles
     * @param string $dir
     * @return array
     * @throws \MicroEngine\Core\Exception\MeException
     */
    public function registerUploadedFiles(array $uploadedFiles, string $dir): array
    {
        $results = [];
        $updated = false;

        $rows = File::where('dir', $dir)->get();
        $records = [];
        foreach ($rows as $row) {
            $records[$row->filename] = $row;
        }

        foreach ($uploadedFiles['files'] as $uploadedFile) {
            $filename = $uploadedFile->getClientFilename();
            $result = [];
            $result['filename'] = $filename;
            $result['class'] = null;

            // エラー確認
            if ($uploadedFile->getError() !== UPLOAD_ERR_OK) {
                $result['message'] = 'エラー';
                $results[] = $result;
                continue;
            }

            // 既存ファイル確認
            if (isset($records[$filename])) {
                $oldFile = $records[$filename];
                $oldHash = $this->fileSystem->md5($oldFile->path());
                $uploadedHash = md5_file($uploadedFile->file);
                if ($oldHash !== $uploadedHash) {
                    // 更新
                    $this->fileSystem->moveUploadedFile($oldFile->path(), $uploadedFile);
                    $this->update($oldFile->id, [
                        'size' => $uploadedFile->getSize(),
                    ], false);
                    $result['message'] = '更新';
                    $result['class'] = 'primary';
                    $updated = true;
                    $results[] = $result;
                } else {
                    // 変更なし
                    $result['message'] = '変更なし';
                    $results[] = $result;
                }
                continue;
            }

            // 新規登録
            $file = $this->create([
                'dir' => $dir,
                'filename' => $filename,
                'extension' => strtolower(pathinfo($filename, PATHINFO_EXTENSION)),
                'size' => $uploadedFile->getSize(),
            ], false);
            $this->fileSystem->moveUploadedFile($file->path(), $uploadedFile);

            $result['message'] = '新規作成';
            $result['class'] = 'success';
            $updated = true;
            $results[] = $result;
        }

        if ($updated) {
            $this->touchDir($dir);
        }
        return $results;
    }

    /**
     * ディレクトリ（files テーブルのレコード）の更新日時を更新する
     *
     * @param string $dir
     */
    protected function touchDir(string $dir)
    {
        $file = File::where('dir', $dir)->first();
        if ($file) {
            $file->touch();
        }
    }

    /**
     * ファイル更新時に旧リサイズ画像を削除する
     *
     * @param File $file
     * @throws \MicroEngine\Core\Exception\MeException
     */
    protected function destroyOldResize(File $file)
    {
        // 旧リサイズファイルを削除
        $options = array_merge(config('image.old_options'), config('image.options'));
        foreach (array_keys($options) as $option) {
            $this->fileSystem->delete(resize($file->path(), $option));
        }
    }
}
