<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Managers;

use MicroEngine\Cms\Models\Page;
use MicroEngine\Cms\Models\Category;
use MicroEngine\Cms\Models\PageRevision;
use Interop\Container\ContainerInterface;
use MicroEngine\Cms\Services\ContentService;
use Illuminate\Database\Capsule\Manager as DB;

class PageManager
{
    use RevisionTrait;

    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var string モデルクラス名
     */
    protected $modelClass = 'MicroEngine\Cms\Models\Page';
    /**
     * @var string IDカラム名
     */
    protected $modelId = 'page_id';
    /**
     * @var string リビジョンクラス名
     */
    protected $revisionClass = 'MicroEngine\Cms\Models\PageRevision';
    /**
     * @var string リビジョンテーブル名
     */
    protected $revisionTable = 'page_revisions';

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
    }

    /**
     * ページ保存
     *
     * @param array $data
     * @return Page
     */
    public function create(array $data): Page
    {
        // パスを組み立てる
        $data['path'] = $this->composePath($data);
        // 作成者・更新者にログインユーザーのIDをセットする
        $data['create_user_id'] = $data['update_user_id'] = $this->c->auth->getUserId();

        $page = Page::create($data);

        $contentService = $this->c->contentService;
        $contentService->saveMeta($page);
        $contentService->saveSearch($page);

        // リビジョン保存
        $revision = $this->saveRevision($page->id, $data);

        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            // 下書きとして保存
            // リビジョンIDを下書きIDとしてセット
            $page->draft_id = $revision->id;
            // 新規で下書き保存した場合には、本稿が存在しないので、下書きだけというステータスを入れる
            $page->visibility = ContentService::VISIBILITY_DRAFT;
            $page->save();
        }

        if (isset($data['category_id'])) {
            $categoryManager = $this->c->categoryManager;
            $categoryManager->updateHasIndex($data['category_id']);
        }

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        // ログを書き出す
        $this->c->logger->addInfo('ページを作成しました。', [
            'type' => $page->type,
            'path' => $page->path,
            'name' => $page->name,
        ]);
        return $page;
    }

    /**
     * ページ更新
     *
     * @param int $id
     * @param array $data
     * @return Page
     */
    public function update(int $id, array $data): Page
    {
        $data['path'] = $this->composePath($data);
        // 更新者にログインユーザーのIDをセットする
        $data['update_user_id'] = $this->c->auth->getUserId();

        // DBからページを取得
        $page = Page::withoutGlobalScopes()->findOrFail($id);
        // リビジョン保存
        $revision = $this->saveRevision($id, $data);

        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            // 下書きとして保存
            // リビジョンIDを下書きIDとしてセット
            $page->draft_id = $revision->id;
            $page->save();
        } else {
            // 本稿として保存
            $oldCategoryId = $page->category_id;
            $page->fill($data);
            // 下書き編集画面から本稿登録の場合は下書きを削除する
            if ($data['edit_mode'] === 'draft') {
                $page->draft_id = null;
            }
            $page->save();

            $contentService = $this->c->contentService;
            $contentService->saveMeta($page);
            $contentService->saveSearch($page);

            $categoryManager = $this->c->categoryManager;
            $categoryManager->updateHasIndex($data['category_id']);

            if ($data['category_id'] != $oldCategoryId) {
                $categoryManager->updateHasIndex($oldCategoryId);
            }
        }

        // ログを書き出す
        if (isset($data['revision_type']) && $data['revision_type'] === 'draft') {
            $message = 'ページを下書きとして保存しました。';
        } else {
            $message = 'ページを更新しました。';
        }
        $this->c->logger->addInfo($message, [
            'type' => $page->type,
            'path' => $page->path,
            'name' => $page->name,
        ]);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $page;
    }

    /**
     * ページ削除
     *
     * @param int $id
     * @param int|null $revisionId
     * @return Page
     */
    public function destroy(int $id, int $revisionId = null): Page
    {
        $page = Page::withoutGlobalScopes()->findOrFail($id);
        // 更新者にログインユーザーのIDをセットする
        $page->update_user_id = $this->c->auth->getUserId();

        // 履歴から再編集している場合。ただし、新規下書き保存状態の場合をのぞく。
        if ($revisionId && $page->visibility != ContentService::VISIBILITY_DRAFT) {
            // 履歴から再編集している場合は、履歴を削除する。
            $revision = PageRevision::findOrFail($revisionId);
            $revision->delete();
            // 削除対象が最新下書きの場合はnullをセットする。
            if ($page->draft_id == $revisionId) {
                $page->draft_id = null;
                $page->save();
                $message = 'ページの下書きを削除しました。';
            } else {
                $message = 'ページの履歴データを削除しました。';
            }

            // ログを書き出す
            $this->c->logger->addInfo($message, [
                'type' => $page->type,
                'path' => $page->path,
                'name' => $page->name,
                'revision_id' => $revisionId,
            ]);
        } else {
            // 履歴から編集していない場合は、本レコードを論理削除する。
            $page->visibility = ContentService::VISIBILITY_TRASH;
            $page->update();
            // カテゴリのインデックスを更新
            $categoryManager = $this->c->categoryManager;
            $categoryManager->updateHasIndex($page->category_id);

            // ログを書き出す
            $this->c->logger->addInfo('ページをゴミ箱に移動しました。', [
                'type' => $page->type,
                'path' => $page->path,
                'name' => $page->name,
            ]);
        }
        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $page;
    }

    /**
     * ページをごみ箱から削除
     *
     * @param int $id
     * @return Page
     */
    public function forceDestroy(int $id): Page
    {
        $page = Page::withoutGlobalScopes()->findOrFail($id);

        // ゴミ箱から削除する場合は、物理削除を行う
        $page->delete();
        // カテゴリのインデックスを更新
        $categoryManager = $this->c->categoryManager;
        $categoryManager->updateHasIndex($page->category_id);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        // ログを書き出す
        $this->c->logger->addInfo('ページをごみ箱から削除しました。', [
            'type' => $page->type,
            'path' => $page->path,
            'name' => $page->name,
        ]);
        return $page;
    }

    /**
     * ごみ箱を空にする
     *
     * @return int
     */
    public function emptyTrash(): int
    {
        $count = 0;
        // ごみ箱内の全コンテンツに対して処理する
        foreach (Page::trashed()->get() as $page) {
            // コンテンツ一つずつをごみ箱から削除する
            $this->forceDestroy($page->id);
            $count++;
        }
        $this->c->logger->addInfo('ページのごみ箱空にしました。', [
            'count' => $count,
        ]);
        return $count;
    }

    /**
     * 一括作成
     *
     * @param array $data
     * @return int $created 作成数
     */
    public function bulkCreate(array $data): int
    {
        $created = 0;
        // idがキーでdirが値の配列を取得
        $dirs = Category::pluck('dir', 'id');

        foreach ($data as $row) {
            // category_dir がある場合は該当するカテゴリIDを category_id にセットする
            if (isset($row['category_dir'])) {
                $categoryDir = $row['category_dir'];
                if ($dirs->contains($categoryDir)) {
                    $row['category_id'] = $dirs->search($categoryDir);
                } else {
                    $category = Category::where('dir', $categoryDir)->first();
                    if ($category !== null) {
                        $row['category_id'] = $category->id;
                        $dirs[$category->id] = $categoryDir;
                    }
                }
            }
            if (isset($row['fields']) && !is_array($row['fields'])) {
                $row['fields'] = json_decode($row['fields'], true);
            }
            $this->create($row);
            $created++;
        }
        $this->c->logger->addInfo('ページを一括作成しました。', [
            'count' => $created,
        ]);
        return $created;
    }

    /**
     * ディレクトリ一括更新
     *
     * @param int $categoryId
     * @param string $dir
     * @return int $updated 更新行数
     */
    public function updateDir(int $categoryId, string $dir): int
    {
        $updated = 0;
        $updated += DB::update('update pages set path = CONCAT(?, filename, ?) where category_id = ? and filename != ?',
            [$dir, HTML_EXTENSION, $categoryId, INDEX_FILENAME]);
        $updated += DB::update('update pages set path = ? where category_id = ? and filename = ?',
            [$dir, $categoryId, INDEX_FILENAME]);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('ページのディレクトリを一括更新しました。', [
            'count' => $updated,
        ]);
        return $updated;
    }

    /**
     * ファイル名の初期値取得
     *
     * @param array $config
     * @return string
     */
    public function initialFilename(array $config): string
    {
        if (!isset($config['inputs']['filename']['value'])) {
            return '';
        }
        // 初期値をセット
        $filename = $config['inputs']['filename']['value'];
        // { が含まれていれば、日時の置換処理を行う。
        if (strpos($filename, '{') !== false) {
            $filename = str_replace('{year}', date('Y'), $filename);
            $filename = str_replace('{month}', date('m'), $filename);
            $filename = str_replace('{day}', date('d'), $filename);
            $filename = str_replace('{hour}', date('H'), $filename);
            $filename = str_replace('{minute}', date('i'), $filename);
            $filename = str_replace('{second}', date('s'), $filename);
        }

        // ファイル名の初期値とdirが設定されている場合は、ユニークなファイル名を生成する
        if (isset($config['dir'])) {
            // dirの値を元にカテゴリIDを取得
            $categoryId = Category::where('dir', $config['dir'])->value('id');
            // ユニークファイル名を取得
            $filename = $this->uniqueFilename($filename, $categoryId);
        }

        return $filename;
    }

    /**
     * ユニークファイル名を取得
     *
     * @param string $initFilename
     * @param int|null $categoryId
     * @return string
     */
    protected function uniqueFilename(string $initFilename, int $categoryId = null): string
    {
        $filename = $initFilename;
        $filenames = Page::withoutGlobalScopes()->where('category_id', $categoryId)->pluck('filename')->toArray();
        $i = 1;
        while (array_search($filename, $filenames) !== false) {
            $filename = $initFilename . '_' . $i++;
        }
        return $filename;
    }

    /**
     * ディレクトリ・ファイル名からパスを組み立てる
     *
     * @param array $data
     * @return string
     */
    protected function composePath(array $data): string
    {
        $category = Category::find($data['category_id']);

        $path = $category->dir;
        // indexファイル以外はファイル名・拡張子を付与する
        if ($data['filename'] !== INDEX_FILENAME) {
            $path .= $data['filename'] . HTML_EXTENSION;
        }
        return $path;
    }
}
