<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Managers;

use MicroEngine\Cms\Models\Table;
use MicroEngine\Cms\Models\tableRevision;
use Interop\Container\ContainerInterface;

class TableManager
{
    use RevisionTrait;

    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var string モデルクラス名
     */
    protected $modelClass = 'MicroEngine\Cms\Models\Table';
    /**
     * @var string IDカラム名
     */
    protected $modelId = 'table_id';
    /**
     * @var string リビジョンクラス名
     */
    protected $revisionClass = 'MicroEngine\Cms\Models\TableRevision';
    /**
     * @var string リビジョンテーブル名
     */
    protected $revisionTable = 'table_revisions';

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
    }

    /**
     * テーブルコンテンツ保存
     *
     * @param array $data
     * @param array $config
     * @return Table
     */
    public function create(array $data, array $config): Table
    {
        // 作成者・更新者にログインユーザーのIDをセットする
        $data['create_user_id'] = $data['update_user_id'] = $this->c->auth->getUserId();

        // 配列の構造調整
        $data = $this->unwrapData($data, $config);
        $table = Table::create($data);

        // リビジョン保存
        $this->saveRevision($table->id, $data);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('テーブルコンテンツを作成しました。', [
            'type' => $table->type,
            'line_count' => $table->line_count,
        ]);
        return $table;
    }

    /**
     * テーブルコンテンツ更新
     *
     * @param int $id
     * @param array $data
     * @param array $config
     * @return Table
     */
    public function update(int $id, array $data, array $config): Table
    {
        // 更新者にログインユーザーのIDをセットする
        $data['update_user_id'] = $this->c->auth->getUserId();

        // 配列の構造調整
        $data = $this->unwrapData($data, $config);

        $table = Table::findOrFail($id);
        $table->fill($data);
        $table->save();

        // リビジョン保存
        $this->saveRevision($id, $data);

        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        $this->c->logger->addInfo('テーブルコンテンツを変更しました。', [
            'type' => $table->type,
            'line_count' => $table->line_count,
        ]);
        return $table;
    }

    /**
     * テーブルコンテンツ削除
     *
     * @param int $id
     * @param int|null $revisionId
     * @return Table
     */
    public function destroy(int $id, int $revisionId = null): Table
    {
        $table = Table::findOrFail($id);
        // 更新者にログインユーザーのIDをセットする
        $table->update_user_id = $this->c->auth->getUserId();

        if ($revisionId) {
            // 履歴から再編集している場合は、履歴を削除する。
            $revision = TableRevision::findOrFail($revisionId);
            $revision->delete();
            // ログを書き出す
            $this->c->logger->addInfo('テーブルコンテンツの履歴を削除しました。', [
                'type' => $table->type,
                'line_count' => $table->line_count,
                'revision_id' => $revisionId,
            ]);
        } else {
            // 履歴から編集していない場合は、本レコードを削除する。
            $table->delete();

            // ログを書き出す
            $this->c->logger->addInfo('テーブルコンテンツを削除しました。', [
                'type' => $table->type,
                'line_count' => $table->line_count,
            ]);
        }
        // キャッシュをクリアする
        $this->c->htmlCache->clear();

        return $table;
    }

    /**
     * 一括作成
     *
     * @param array $data
     * @return int $created 作成数
     */
    public function bulkCreate(array $data): int
    {
        $created = 0;

        foreach ($data as $row) {
            if (isset($row['data']) && !is_array($row['data'])) {
                $row['data'] = json_decode($row['data'], true);
            }
            // 該当タイプの設定取得
            $config = config('tables.' . $row['type']);

            $this->create($row, $config);
            $created++;
        }
        $this->c->logger->addInfo('テーブルコンテンツを一括作成しました。', [
            'count' => $created,
        ]);
        return $created;
    }

    /**
     * 配列の構造調整
     *
     * @param array $data
     * @param array $config
     * @return array
     */
    public function unwrapData(array $data, array $config): array
    {
        $dataType = $config['data_type'];

        if (!isset($data['data'])) {
            $data['line_count'] = 0;
            return $data;
        }
        if (is_string($data['data'])) {
            $rows = json_decode($data['data'], true);
        } else {
            $rows = $data['data'];
        }

        if ($dataType === 'list') {
            // 包んでいる配列を取り除く
            $list = [];
            foreach ($rows as $row) {
                $list[] = $row[0];
            }
            // 空行削除
            $data['data'] = array_values(array_filter($list, 'strlen'));
        } elseif ($dataType === 'hash') {
            // 包んでいる配列を取り除く
            $hash = [];
            foreach ($rows as $key => $row) {
                if (is_array($row)) {
                    $hash[$row[0]] = $row[1];
                } else {
                    $hash[$key] = $row;
                }
            }
            // 空行削除
            $data['data'] = array_filter($hash, 'strlen', ARRAY_FILTER_USE_KEY);
        } elseif ($dataType === 'matrix') {
            // クライアントサイドでは、配列の配列で扱われるので、配列の連想配列に変換する
            $matrix = [];
            foreach ($rows as $row) {
                $hash = [];
                foreach (array_keys($config['columns']) as $colKey) {
                    $hash[$colKey] = array_shift($row);
                }
                // 空行削除
                if (!strlen(implode($hash))) {
                    continue;
                }
                $matrix[] = $hash;
            }
            $data['data'] = $matrix;
        }
        $data['line_count'] = count($data['data']);

        return $data;
    }
}
