<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Cms\Services;

use MicroEngine\Cms\Models\Page;
use MicroEngine\Cms\Models\Module;
use MicroEngine\Cms\Models\Calendar;
use MicroEngine\Cms\Models\Collection;
use Interop\Container\ContainerInterface;
use Illuminate\Database\Eloquent\Builder;
use MicroEngine\Core\Exception\MeException;
use MicroEngine\Core\Pagination\Pagination;
use Illuminate\Database\Eloquent\Collection as EloquentCollection;

class Query
{
    /**
     * sql の limit の最大値
     */
    const LIMIT_MAX_SIZE = 2147483647;
    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var string ベースタイプ
     */
    protected $baseType;
    /**
     * @var string テーブル名
     */
    protected $table;

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
    }

    /**
     * 設定に定義した条件でコンテンツリストを取得する
     *
     * @param string $key
     * @param array $params
     * @return EloquentCollection
     * @throws \Exception
     */
    public function query(string $key, array $params = []): EloquentCollection
    {
        // クエリ設定を取得
        $config = $this->queryConfig($key, $params);

        // クエリ設定を元にコンテンツを取得
        if (isset($config['class'])) {
            // classが指定されていれば、Appのクラスを実行する
            $class = 'App\\Queries\\' . ucfirst($config['class']);
            $obj = new $class($this->c);
            $content = $obj($config, $params);
        } else {
            $content = $this->get($config);
        }
        return $content;
    }

    /**
     * コンテンツリストを取得する
     *
     * @param  array $config
     * @return EloquentCollection
     * @throws \Exception
     */
    public function get(array $config = []): EloquentCollection
    {
        // クエリを組み立てる
        $query = $this->buildQuery($config);

        // 対象コンテンツリストを取得する
        $items = $query->get([$this->table . '.*']);

        //$sql = $query->toSql();
        return $items;
    }

    /**
     * ページネーションクエリ
     *
     * @param string $key
     * @param $pageNo
     * @param array $paginationOptions
     * @param array $params
     * @return Pagination
     * @throws \Exception
     */
    public function pagination(string $key, $pageNo, $paginationOptions = [], array $params = []): Pagination
    {
        // クエリ設定を取得
        $config = $this->queryConfig($key, $params);

        // クエリを組み立てる
        $query = $this->buildQuery($config);

        return $query->pagination($pageNo, $paginationOptions);
    }

    /**
     * クエリ設定を取得
     *
     * @param string $key
     * @param array $params
     * @return array
     */
    protected function queryConfig(string $key, array $params = []): array
    {
        $queryConfig = config('query');
        $config = $queryConfig[$key];
        $config += [
            'base_type' => 'page'
        ];

        // 配列以外の値が渡された場合は param と置換する
        if ($params !== null && !is_array($params)) {
            $params = ['param' => $params];
        }

        // プレースホルダをパラメータで置換する。
        if (!empty($params)) {
            array_walk_recursive($config, function (&$item, $k, $params) {
                foreach ($params as $search => $value) {
                    $item = str_replace('{' . $search . '}', $value, $item);
                }
            }, $params);
        }
        return $config;
    }

    /**
     * クエリを組み立てる
     *
     * @param array $config
     * @return Builder
     * @throws \Exception
     */
    protected function buildQuery(array $config = []): Builder
    {
        // 初期設定値とオプションをマージする
        $config += [
            'base_type' => null,
            'type' => null,
            'category' => null,
            'dir' => null,
            'filename' => null,
            'exclude_filename' => [],
            'join' => [],
            'search' => null,
            'where' => [],
            'meta' => [],
            'order' => 'display_order',
            'sort' => 'asc',
            'extra_order' => [],
            'offset' => null,
            'limit' => null,
        ];
        if (empty($config['base_type'])) {
            $config['base_type'] = $this->findBaseType($config['type']) ?? 'page';
        }
        $this->baseType = $config['base_type'];
        $this->table = $this->baseType . 's';

        $query = $this->getQuery($this->baseType);
        $this->join($query, $config['join']);
        $this->type($query, $config['type']);
        $this->category($query, $config['category']);
        $this->dir($query, $config['dir']);
        $this->filename($query, $config['filename']);
        $this->excludeFilename($query, $config['exclude_filename']);
        $this->meta($query, $config['meta']);
        $this->where($query, $config['where']);
        $this->search($query, $config);
        $this->orderBy($query, $config);
        $this->limit($query, $config['limit']);
        $this->offset($query, $config);

        return $query;
    }

    /**
     * タイプを元にベースタイプを返す
     *
     * @param string $type
     * @return string|null
     * @throws \Exception
     */
    protected function findBaseType(string $type)
    {
        $baseTypes = ['page', 'collection', 'module', 'calendar'];

        // 各設定ファイルに指定されたタイプがあるか確認する
        foreach ($baseTypes as $baseType) {
            $config = config($baseType . 's.' . $type);
            if ($config) {
                return $baseType;
            }
        }
        // デバッグモードの場合は例外を発生させる
        if (settings('debug', false)) {
            throw new MeException('コンテンツ設定が見つかりません。');
        }
        return null;
    }

    /**
     * クエリオブジェクトを取得
     *
     * @param string $baseType
     * @return Builder
     */
    protected function getQuery(string $baseType): Builder
    {
        // 対象ベースタイプの取得クエリを作成する
        if ($baseType === 'collection') {
            $query = Collection::query();
        } elseif ($baseType === 'calendar') {
            $query = Calendar::query();
        } elseif ($baseType === 'module') {
            $query = Module::query();
        } else {
            $query = Page::query();
        }
        return $query;
    }

    /**
     * タイプ検索
     *
     * @param Builder $query
     * @param string $setting
     */
    protected function type(Builder $query, string $setting = null)
    {
        // ページタイプ指定
        if ($setting !== null) {
            $query->where('type', $setting);
        }
    }

    /**
     * カテゴリー検索
     *
     * @param Builder $query
     * @param null $setting
     */
    protected function category(Builder $query, $setting = null)
    {
        if ($setting === null) {
            return;
        }
        if (!is_array($setting) && is_numeric($setting)) {
            $setting = [$setting];
        }

        // カテゴリ指定
        $query->whereIn('category_id', $setting);
    }

    /**
     * ディレクトリ検索
     *
     * @param Builder $query
     * @param string $setting
     */
    protected function dir(Builder $query, string $setting = null)
    {
        // ディレクトリ指定
        if ($setting !== null) {
            $dir = end_slash($setting);
            $query->where('path', 'like', $dir . '%');
        }
    }

    /**
     * ファイル名検索
     *
     * @param Builder $query
     * @param $setting
     */
    protected function filename(Builder $query, $setting)
    {
        // ファイル名指定
        if ($setting === null) {
            return;
        }
        if (is_array($setting)) {
            $query->whereIn('filename', $setting);
        } else {
            $query->where('filename', $setting);
        }
    }

    /**
     * 除外ファイル名指定
     *
     * @param Builder $query
     * @param $setting
     */
    protected function excludeFilename(Builder $query, $setting)
    {
        if (empty($setting)) {
            return;
        }
        if (is_array($setting)) {
            $query->whereNotIn('filename', $setting);
        } else {
            $query->where('filename', '!=', $setting);
        }
    }

    /**
     * メタデータ絞り込み
     *
     * @param Builder $query
     * @param array $setting
     */
    protected function meta(Builder $query, array $setting)
    {
        if (empty($setting) || !is_array($setting)) {
            return;
        }
        $query->whereHas('metas', function ($query) use ($setting) {
            $i = 1;
            $metaSetting = [];
            foreach ($setting as $item) {
                $metaSetting['or_and.' . $i++] = [
                    ['name', $item[0]],
                    ['value', $item[1]],
                ];
            }
            $this->where($query, $metaSetting);
        });
    }

    /**
     * メタテーブルJOIN
     *
     * @param Builder $query
     * @param array $setting
     */
    protected function join(Builder $query, array $setting)
    {
        if (empty($setting) || !is_array($setting)) {
            return;
        }
        foreach ($setting as $alias) {
            $query->join($this->baseType . '_metas as ' . $alias, function ($join) use ($alias) {
                $join->on($this->table . '.id', '=', $alias . '.' . $this->baseType . '_id')->where($alias . '.name', $alias);
            });
        }
    }

    /**
     * where 指定
     *
     * @param Builder $query
     * @param array $setting
     */
    protected function where(Builder $query, array $setting)
    {
        if (empty($setting) || !is_array($setting)) {
            return;
        }

        // type や 公開条件 と or 条件になることを避けるために、クロージャを挟む。
        // クロージャを挟むとsqlに変換されたときに、これらのwhere条件全体が () で囲われる。
        $query->where(function ($query) use ($setting) {
            $this->whereCondition($query, $setting);
        });
    }

    /**
     * where 条件を構築
     *
     * @param Builder $query
     * @param array $conditions
     */
    protected function whereCondition(Builder $query, array $conditions)
    {
        // 論理演算子の初期値
        $logicalOperator = 'and';

        foreach ($conditions as $condition) {
            if (is_string($condition)) {
                // 文字列の場合は、演算子として採用する
                $logicalOperator = $condition;
            } else if (is_array($condition)) {
                // 配列の場合
                if (isset($condition[0]) && is_array($condition[0])) {
                    // 配列の１つ目の要素が配列の場合はネストさせるためにクロージャを使う
                    if (strcasecmp($logicalOperator, 'or') === 0) {
                        $query->orWhere(function ($query) use ($condition) {
                            $this->whereCondition($query, $condition);
                        });
                    } else {
                        $query->where(function ($query) use ($condition) {
                            $this->whereCondition($query, $condition);
                        });
                    }
                } else {
                    list($c, $o, $v) = $this->whereParams($condition);
                    if (strcasecmp($logicalOperator, 'or') === 0) {
                        $query->orWhere($c, $o, $v);
                    } else {
                        $query->where($c, $o, $v);
                    }
                }
                // 論理演算子を初期値に戻す
                $logicalOperator = 'and';
            }
        }
    }

    /**
     * where 用配列のパラメータ数を調整する
     *
     * @param array $where
     * @return array
     */
    protected function whereParams(array $where): array
    {
        $ret = [];

        if (count($where) === 2) {
            // where配列が2個の場合は、演算子はデフォルトの = を指定する。
            $ret[0] = $where[0];
            $ret[1] = '=';
            $ret[2] = $where[1];
        } else {
            // where配列が2個以外の場合はそのまま返す
            $ret = $where;
        }
        return $ret;
    }

    /**
     * 検索クエリ
     *
     * @param Builder $query
     * @param array $config
     */
    protected function search(Builder $query, array $config)
    {
        $search = $config['search'] ?? null;
        if ($search === null) {
            return;
        }
        $query->join($this->baseType . '_searches as fts', function ($join) {
            $join->on($this->table . '.id', '=', 'fts.' . $this->baseType . '_id');
        });

        $searchMode = $config['search_mode'] ?? 'like_and';
        if ($searchMode === 'fulltext') {
            $this->fullTextSearch($query, $search);
        } else if ($searchMode === 'like_or') {
            $this->likeSearch($query, $search, 'or');
        } else {
            $this->likeSearch($query, $search, 'and');
        }
    }

    /**
     * like検索
     *
     * @param Builder $query
     * @param null $searchWord
     * @param string $operator
     */
    protected function likeSearch(Builder $query, $searchWord = null, $operator = 'and')
    {
        // 半角空白等と全角空白で分離する
        $words = preg_split('/[\s]+/u', $searchWord);

        $conditions = [];
        $i = 0;
        foreach ($words as $word) {
            if (!strlen($word)) {
                // 空文字はスルー
                continue;
            }
            if ($i !== 0 && $operator === 'or') {
                // or （like_or）が指定されている場合は、２つ目以降の前に or を挟む
                $conditions[] = 'or';
            }
            $conditions[] = ['fts.content', 'like', '%' . $word . '%'];
            // 配列のキーを使うと、空文字のスルー時に問題があるので、別途カウントする
            $i++;
        }
        // where条件を追加
        $this->where($query, $conditions);
    }

    /**
     * 全文検索
     *
     * @param Builder $query
     * @param null $searchWord
     */
    protected function fullTextSearch(Builder $query, $searchWord = null)
    {
        // 半角空白等と全角空白で分離する
        $words = preg_split('/[\s]+/u', $searchWord);

        $conditions = [];
        foreach ($words as $word) {
            $length = mb_strlen($word, 'UTF-8');
            if ($length === 0) {
                // 空文字はスルー
                continue;
            }
            if ($length === 1) {
                // N-gram で２文字でインデックスを作っているので、１文字に一致させるために * を追加する
                $conditions[] = $word . '*';
            } else {
                $conditions[] = $word;
            }
        }
        $adjustedWord = implode(' ', $conditions);
        $query->whereRaw("match(`fts`.`content`) against (? IN BOOLEAN MODE)", [$adjustedWord]);
    }

    /**
     * order by を指定
     *
     * @param Builder $query
     * @param array $config
     */
    protected function orderBy(Builder $query, array $config)
    {
        // 並び順対応
        $query->orderBy($config['order'], $config['sort']);
        if (empty($config['extra_order']) || !is_array($config['extra_order'])) {
            return;
        }
        // extra_order 設定に配列の値がある場合は、order by を追加指定する
        foreach ($config['extra_order'] as $key => $value) {
            $query->orderBy($key, $value);
        }
    }

    /**
     * limit を指定
     *
     * @param Builder $query
     * @param $setting
     */
    protected function limit(Builder $query, $setting)
    {
        // limitを指定
        if ($setting !== null) {
            $query->take($setting);
        }
    }

    /**
     * offset を指定
     *
     * @param Builder $query
     * @param $config
     */
    protected function offset(Builder $query, $config)
    {
        // offsetとlimitを指定
        if ($config['offset'] === null) {
            return;
        }
        if ($config['limit'] === null) {
            // offset のみ指定されている場合は、limit も指定する。
            $query->take(self::LIMIT_MAX_SIZE);
        }
        $query->skip($config['offset']);
    }
}
