<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Core;

use Monolog\Logger;
use Slim\Csrf\Guard;
use Slim\Flash\Messages;
use Slim\Views\TwigExtension;
use Monolog\Handler\StreamHandler;
use Monolog\Processor\WebProcessor;
use MicroEngine\Core\License\License;
use MicroEngine\Core\Log\UserProcessor;
use MicroEngine\Core\Mail\MailComposer;
use Illuminate\Database\Capsule\Manager;
use MicroEngine\Core\Providers\MailerProvider;
use MicroEngine\Core\Middleware\InputValidationMiddleware;
use MicroEngine\Core\Views\{
    View, MenuManager, CoreExtension
};
use MicroEngine\Core\Handler\{
    Error, NotAllowed, NotFound, PhpError
};
use MicroEngine\Core\Support\{
    Date, ServiceProvider, Util, Base64Stream
};

class CoreServiceProvider extends ServiceProvider
{
    /**
     * サービス登録
     */
    public function register()
    {
        $c = $this->c;

        $c['errorHandler'] = function ($c) {
            return new Error(config('settings.displayErrorDetails', false));
        };
        $c['phpErrorHandler'] = function ($c) {
            return new PhpError(config('settings.displayErrorDetails', false));
        };
        $c['notAllowedHandler'] = function ($c) {
            return new NotAllowed();
        };
        $c['notFoundHandler'] = function ($c) {
            return new NotFound($c);
        };
        $c['db'] = function ($c) {
            $capsule = new Manager;
            $connection = config('db.default', 'mysql');
            $capsule->addConnection(config('db.connections.' . $connection));
            $capsule->setAsGlobal();
            $capsule->bootEloquent();

            return $capsule;
        };
        $c['csrf'] = function ($c) {
            $guard = new Guard();
            $guard->setPersistentTokenMode(true);
            return $guard;
        };
        $c['flash'] = function () {
            return new Messages();
        };
        $c['view'] = function ($c) {
            $view = new View(__DIR__ . '/../../app/themes', [
                'cache' => __DIR__ . '/../../storage/cache/views',
                'auto_reload' => true,
                'debug' => config('settings.debug', false),
            ]);
            $loader = $view->getLoader();
            $loader->addPath(themes_root());
            $loader->addPath(__DIR__ . '/../../app/resources', 'app');
            $basePath = rtrim(str_ireplace('index.php', '', $c['request']->getUri()->getBasePath()), '/');
            $view->addExtension(new TwigExtension($c['router'], $basePath));
            $view->addExtension(new CoreExtension($c));
            $view->addExtension(new \Twig_Extension_StringLoader());
            $view->addExtension(new \Twig_Extension_Debug());

            return $view;
        };
        $c['menuManager'] = function ($c) {
            return new MenuManager($c);
        };
        $c['date'] = function () {
            return new Date();
        };
        $c['util'] = function ($c) {
            return new Util($c);
        };
        $c['logger'] = function ($c) {
            $level = config('settings.logger.level');
            $logger = new Logger('App');
            $logger->pushProcessor(new WebProcessor(null, ['ip', 'http_method', 'url']));
            $logger->pushProcessor(new UserProcessor($c));
            $filename = current_log_filename();
            $logger->pushHandler(new StreamHandler(system_root('storage/logs/' . $filename), $level));

            return $logger;
        };
        $c['license'] = function ($c) {
            Base64Stream::register();
            require_once('base64://' . __DIR__ . '/License/License.php');

            return new License($c);
        };
        $c->register(new MailerProvider());
        $c['mailComposer'] = function ($c) {
            return new MailComposer($c->view, $c->mailer);
        };
    }

    /**
     * サービス起動
     */
    public function boot()
    {
        // モデルを静的に実行できるように、コネクションを登録する。ここでは呼び出すだけ
        $db = $this->c->get('db');

        // アプリケーションミドルウェアに登録
        $this->app->add($this->c->csrf);
        $this->app->add(new InputValidationMiddleware($this->c));
    }
}
