<?php
/**
 * MicroEngine CMS
 * https://microengine.jp
 *
 * (c) 2018 MicroEngine Inc.
 */

namespace MicroEngine\Core\Views;

use Twig_Filter;
use Twig_Function;
use Twig_Extension;
use Interop\Container\ContainerInterface;

class CoreExtension extends Twig_Extension
{
    /**
     * @var ContainerInterface
     */
    protected $c;
    /**
     * @var MenuManager
     */
    protected $menuManager;

    /**
     * @param ContainerInterface $c
     */
    public function __construct(ContainerInterface $c)
    {
        $this->c = $c;
        $this->menuManager = $c->menuManager;
    }

    /**
     * 関数登録
     *
     * @return array
     */
    public function getFunctions(): array
    {
        return [
            new Twig_Function('csrf', [$this, 'csrf']),
            new Twig_Function('flash', function () {
                return $this->c->flash;
            }),
            new Twig_Function('theme', 'theme'),
            new Twig_Function('themes', 'themes'),
            new Twig_Function('current_menu', function ($main, $sub = null) {
                return $this->menuManager->setCurrent($main, $sub);
            }),
            new Twig_Function('main_menu', function () {
                return $this->menuManager->mainMenu();
            }),
            new Twig_Function('sub_menu', function () {
                return $this->menuManager->subMenu();
            }),
            new Twig_Function('config', 'config'),
            new Twig_Function('history_has', 'history_has'),
            new Twig_Function('history', 'history'),
            new Twig_Function('file_type_icon', 'file_type_icon'),
        ];
    }

    /**
     * フィルター登録
     *
     * @return array
     */
    public function getFilters(): array
    {
        return [
            new Twig_Filter('filesize_format', [$this, 'filesizeFormat']),
            new Twig_Filter('basename', 'basename'),
            new Twig_Filter('recent_updates', function ($time, $today = 'today', $fresh = 'fresh') {
                return $this->c->date->recentUpdates($time, $today, $fresh);
            }),
            new Twig_Filter('date_local', function ($time, $format) {
                return $this->c->date->localFormat($time, $format);
            }),
            new Twig_Filter('date_admin', function ($time) {
                return $this->c->date->formatName($time, 'date_admin');
            }),
            new Twig_Filter('date_long', function ($time) {
                return $this->c->date->formatName($time, 'date_long');
            }),
            new Twig_Filter('date_short', function ($time) {
                return $this->c->date->formatName($time, 'date_short');
            }),
            new Twig_Filter('date_local_long', function ($time) {
                return $this->c->date->localFormatName($time, 'date_local_long');
            }),
            new Twig_Filter('date_local_short', function ($time) {
                return $this->c->date->localFormatName($time, 'date_local_short');
            }),
            new Twig_Filter('then', [$this, 'then']),
        ];
    }

    /**
     * ファイルサイズの単位を変換する
     *
     * @param $bytes
     * @return string
     */
    public function filesizeFormat($bytes)
    {
        if (!is_numeric($bytes)) {
            return $bytes;
        }

        $units = array('バイト', 'KB', 'MB', 'GB', 'TB', 'PB');
        if ($bytes > 0) {
            $unit = intval(log($bytes, 1024));

            if (array_key_exists($unit, $units) === true) {
                $ret = sprintf('%d %s', $bytes / pow(1024, $unit), $units[$unit]);
            } else {
                $ret = number_format($bytes) . ' ' . $units[0];
            }
        } else {
            $ret = sprintf('%d %s', $bytes, $units[0]);
        }
        return $ret;
    }

    /**
     * 値が空ではない場合に渡されたの文字列を返す
     *
     * @param mixed $value
     * @param string $default
     * @return string
     */
    public function then($value = null, $default = '')
    {
        if (!twig_test_empty($value)) {
            return $default;
        }
        return '';
    }

    /**
     * CSRFトークンを返す
     *
     * @return array
     */
    public function csrf(): array
    {
        $csrf = $this->c->csrf;
        // CSRF token name and value
        $nameKey = $csrf->getTokenNameKey();
        $valueKey = $csrf->getTokenValueKey();
        $name = $csrf->getTokenName();
        $value = $csrf->getTokenValue();

        return [
            'keys' => [
                'name' => $nameKey,
                'value' => $valueKey
            ],
            'name' => $name,
            'value' => $value
        ];
    }
}
